use std::{
	io::{self, Write},
	path::PathBuf,
};

use clap::Parser;
use cli::{CliArgs, FilesArgs, TagArgs, TagsArgs, UntagArgs};
use taggo::{init_db, try_database};

use anyhow::{Result, anyhow};

mod cli;

pub const TERM_RESET: &str = "\x1b[0m";
pub const TERM_BOLD: &str = "\x1b[1m";

fn main() -> anyhow::Result<()> {
	let args = CliArgs::parse();

	match args.commands {
		cli::Commands::Init => init_db().map_err(|_| anyhow!("Database already exists!")),
		cli::Commands::Tag {
			args,
			files,
			file_tags,
		} => handle_tag(args, files, file_tags),
		cli::Commands::Untag(args) => handle_untag(args),
		cli::Commands::Tags(args) => handle_tags(args),
		cli::Commands::Files(args) => handle_files(args),
	}
}

fn handle_files(args: FilesArgs) -> Result<()> {
	let conn = try_database()?;

	let mut files = taggo::list_files(&conn, args.all, args.query)?;

	let mut w = io::stdout();
	if !files.files.is_empty() {
		if args.all {
			writeln!(&mut w, "{TERM_BOLD}Tagged Files: {TERM_RESET}")?;
		}
		for (file, tags) in files.files {
			// Remove already tagged files
			if let Some(idx) = files.paths.iter().position(|v| *v == file) {
				files.paths.swap_remove(idx);
			}

			if files.show_tags {
				writeln!(&mut w, "{}: {}", file, tags)?;
			} else {
				writeln!(&mut w, "{}", file)?;
			}
		}
	} else {
		writeln!(&mut w, "No files registered!")?;
	}

	if args.all && !files.paths.is_empty() {
		writeln!(&mut w, "\n{TERM_BOLD}Untagged Files: {TERM_RESET}")?;
		for path in files.paths {
			let path_str = path;
			writeln!(&mut w, "{}", path_str)?;
		}
	}

	w.flush()?;

	Ok(())
}

pub fn handle_tag(args: TagArgs, files: Vec<PathBuf>, file_tags: Vec<String>) -> Result<()> {
	let mut conn = try_database()?;

	if let Some(file) = args.file {
		taggo::tag_file(&mut conn, &file, &args.tags)?;
	} else {
		// TODO: Maybe rework this at some point so its more optimized for batch insertion ?
		for file in &files {
			taggo::tag_file(&mut conn, file, &file_tags)?;
		}
	}

	Ok(())
}

pub fn handle_untag(args: UntagArgs) -> Result<()> {
	let mut conn = try_database()?;

	taggo::untag_file(&mut conn, args.file, args.tags)?;

	Ok(())
}

pub fn handle_tags(args: TagsArgs) -> Result<()> {
	let mut conn = try_database()?;

	match args.commands {
		Some(cli::TagsCommands::List) | None => {
			let mut w = io::stdout();
			let tags = taggo::list_tags(&conn)?;
			if !tags.is_empty() {
				for tag in tags {
					writeln!(&mut w, "{} - {}", tag.0, tag.1)?;
				}
			} else {
				writeln!(&mut w, "No tags registered!")?;
			}
			w.flush()?;
			Ok(())
		}
		Some(cli::TagsCommands::Add { add }) => taggo::add_tags(&mut conn, add),
		Some(cli::TagsCommands::Remove { remove }) => taggo::remove_tags(&mut conn, remove),
		Some(cli::TagsCommands::Rename { renames }) => taggo::rename_tag(&conn, renames),
	}
}
