use std::time::{Duration, Instant};

use anyhow::Result;
use ratatui::{
	DefaultTerminal,
	buffer::Buffer,
	crossterm::event::{self, KeyCode, KeyEventKind},
	layout::{Constraint::*, Layout, Rect},
	style::{Color, Modifier, Style},
	widgets::{
		Block, BorderType, Borders, HighlightSpacing, List, ListState, StatefulWidget, Widget,
	},
};
use taggo::try_database;

const SELECTED_STYLE: Style = Style::new()
	.bg(Color::DarkGray)
	.add_modifier(Modifier::BOLD);

pub struct App {
	is_running: bool,
	tick: u64,

	files: taggo::Files,
	list_state: ListState,
}

impl App {
	pub fn new() -> Result<Self> {
		let conn = try_database()?;

		let files = taggo::list_files(&conn, true, None)?;

		Ok(Self {
			is_running: true,
			tick: 0,

			files,
			list_state: ListState::default(),
		})
	}

	pub fn run(mut self, mut terminal: DefaultTerminal) -> Result<()> {
		let tick_rate = Duration::from_millis(20);
		let mut last_tick = Instant::now();

		while self.is_running {
			terminal.draw(|frame| frame.render_widget(&mut self, frame.area()))?;

			let timeout = tick_rate.saturating_sub(last_tick.elapsed());
			if event::poll(timeout)? {
				self.handle_events()?;
			}

			if last_tick.elapsed() >= tick_rate {
				self.update();
				last_tick = Instant::now();
			}
		}

		Ok(())
	}

	fn handle_events(&mut self) -> Result<()> {
		if let ratatui::crossterm::event::Event::Key(key) = event::read()? {
			if key.kind == KeyEventKind::Press {
				match key.code {
					KeyCode::Char('j') | KeyCode::Down => self.list_state.select_next(),
					KeyCode::Char('k') | KeyCode::Up => self.list_state.select_previous(),
					KeyCode::Char('q') | KeyCode::Char('Q') | KeyCode::Esc => self.quit(),
					_ => {}
				}
			}
		}

		Ok(())
	}

	fn update(&mut self) {
		self.tick = self.tick.wrapping_add(1);
	}

	fn quit(&mut self) {
		self.is_running = false;
	}

	fn render_files_list(&mut self, area: Rect, buf: &mut Buffer) {
		let block = Block::new()
			.borders(Borders::all())
			.border_type(BorderType::Rounded);

		let items: Vec<String> = self.files.files.iter().map(|f| f.0.clone()).collect();

		let list = List::new(items)
			.block(block)
			.highlight_style(SELECTED_STYLE)
			.highlight_symbol(">")
			.highlight_spacing(HighlightSpacing::Always);

		StatefulWidget::render(list, area, buf, &mut self.list_state);
	}
}

impl Widget for &mut App {
	fn render(self, area: Rect, buf: &mut Buffer) {
		let layout = Layout::horizontal([Percentage(50), Percentage(50)]);
		let [list_area, info_area] = layout.areas(area);

		let block = Block::new()
			.borders(Borders::all())
			.border_type(BorderType::Rounded);

		if let Some(i) = self.list_state.selected() {
			let tags = self.files.files[i].clone().1;
			let tags: Vec<&str> = tags.split(',').collect();
			let list = List::new(tags)
				.block(block)
				.highlight_style(SELECTED_STYLE)
				.highlight_symbol(">")
				.highlight_spacing(HighlightSpacing::Always);
			StatefulWidget::render(list, info_area, buf, &mut ListState::default());
		}

		self.render_files_list(list_area, buf);
	}
}
